from pydantic import BaseModel
from typing import Dict, Any
from blok import blok, InitContext, Option, ExecutionContext
from blok.tree import YamlFile
from arkitekt_next.bloks.services.config import ConfigService


class AdminCredentials(BaseModel):
    password: str
    username: str
    email: str


@blok(ConfigService)
class ConfigBlok:
    def __init__(self) -> None:
        self.config_path = "configs"
        self.registered_configs = {}

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

    def build(self, ex: ExecutionContext):
        for name, file in self.registered_configs.items():
            ex.file_tree.set_nested(*f"{self.config_path}/{name}".split("/"), file)

    def register_config(self, name: str, file: YamlFile) -> str:
        self.registered_configs[name] = file
        return f"./{self.config_path}/" + name

    def get_path(self, name: str) -> str:
        return f"./{self.config_path}/" + name

    def get_options(self):
        config_path = Option(
            subcommand="config_path",
            help="Which path to use for configs",
            default=self.config_path,
            show_default=True,
        )

        return [config_path]



from herre_next.herre import Herre
from fakts_next import Fakts
from herre_next.grants.oauth2.refresh import RefreshGrant
from herre_next.fakts.fakts_endpoint_fetcher import FaktsUserFetcher
from herre_next.fakts.grant import FaktsGrant
from arkitekt_next.base_models import User
from arkitekt_next.apps.service.grant_registry import ARKITEKT_GRANT_REGISTRY


class ArkitektNextHerre(Herre):
    pass


def build_arkitekt_next_herre_next(fakts_next: Fakts) -> ArkitektNextHerre:
    return ArkitektNextHerre(
        grant=RefreshGrant(
            grant=FaktsGrant(
                fakts=fakts_next,
                fakts_group="lok",
                grant_registry=ARKITEKT_GRANT_REGISTRY,
            ),
        ),
        fetcher=FaktsUserFetcher(
            fakts=fakts_next, fakts_key="lok.userinfo_url", userModel=User
        ),
    )



from importlib.metadata import version
from pydantic import BaseModel, Field, field_validator
import datetime
from typing import List, Optional, Union, Literal, Dict
from enum import Enum
import semver
import uuid
from arkitekt_next.base_models import Requirement
from string import Formatter
import os

from kabinet.api.schema import (
    AppImageInput,
    InspectionInput,
    SelectorInput,
    ManifestInput,
)

ALLOWED_BUILDER_KEYS = [
    "tag",
    "dockerfile",
    "package_version",
]


class SelectorType(str, Enum):
    RAM = "ram"
    CPU = "cpu"
    GPU = "gpu"
    LABEL = "label"


class Flavour(BaseModel):
    selectors: List[SelectorInput]
    description: str = Field(default="")
    dockerfile: str = Field(default="Dockerfile")
    build_command: List[str] = Field(
        default_factory=lambda: [
            "docker",
            "build",
            "-t",
            "{tag}",
            "-f",
            "{dockerfile}",
            ".",
        ]
    )

    @field_validator("build_command", mode="before")
    def check_valid_template_name(cls, value):
        """Checks that the build_command templates are valid"""

        for v in value:
            for literal_text, field_name, format_spec, conversion in Formatter().parse(
                v
            ):
                if field_name is not None:
                    assert (
                        field_name in ALLOWED_BUILDER_KEYS
                    ), f"Invalid template key {field_name}. Allowed keys are {ALLOWED_BUILDER_KEYS}"

        return value

    def generate_build_command(self, tag: str, relative_dir: str):
        """Generates the build command for this flavour"""

        dockerfile = os.path.join(relative_dir, self.dockerfile)

        return [v.format(tag=tag, dockerfile=dockerfile) for v in self.build_command]

    def check_relative_paths(self, flavour_folder: str):
        """Checks that the paths are relative to the flavour folder"""

        dockerfile_path = os.path.join(flavour_folder, self.dockerfile)

        if not os.path.exists(dockerfile_path):
            raise Exception(
                f"Could not find Dockerfile {self.dockerfile} in flavour {flavour_folder}"
            )


class DeploymentsConfigFile(BaseModel):
    """The ConfigFile is a pydantic model that represents the deployments.yaml file


    Parameters
    ----------
    BaseModel : _type_
        _description_
    """

    app_images: List[AppImageInput] = []
    latest_app_image: Optional[str] = None


class Build(BaseModel):
    build_run: str
    build_id: str
    inspection: Optional[InspectionInput] = None
    description: str = Field(default="")
    selectors: List[SelectorInput] = Field(default_factory=list)
    flavour: str = Field(default="vanilla")
    manifest: ManifestInput
    build_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    base_docker_command: List[str] = Field(
        default_factory=lambda: ["docker", "run", "-it", "--net", "host"]
    )
    base_arkitekt_next_command: List[str] = Field(
        default_factory=lambda: ["arkitekt-next", "run", "prod", "--headless"]
    )

    def build_docker_command(self) -> List[str]:
        """Builds the docker command for this build"""

        base_command = self.base_docker_command

        for selector in self.selectors:
            base_command = base_command + selector.build_docker_params()

        base_command = base_command + [self.build_id]

        return base_command

    def build_arkitekt_next_command(self, fakts_next_url: str):
        """Builds the arkitekt_next command for this build"""

        base_command = self.base_arkitekt_next_command

        for selector in self.selectors:
            base_command = base_command + selector.build_arkitekt_next_params()

        base_command = base_command + ["--url", fakts_next_url]

        return base_command


class BuildsConfigFile(BaseModel):
    builds: List[Build] = Field(default_factory=list)
    latest_build_run: Optional[str] = None



from typing import Dict, Any
import secrets

from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.services.livekit import LivekitService, LivekitCredentials
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo


@blok(LivekitService, description="a self-hosted livekit server")
class LocalLiveKitBlok:
    def __init__(self) -> None:
        self.host = "livekit"
        self.command = ["--dev", "--bind", "0.0.0.0"]
        self.image = "livekit/livekit-server:latest"
        self.mount_repo = True
        self.build_repo = True
        self.secret_key = secrets.token_hex(16)
        self.ensured_repos = []
        self.port_range = [50000, 50030]
        self.api_key = "devkey"
        self.api_secret = "secret"
        self.skip = False

    def preflight(self, init: InitContext, gateway: GatewayService):

        for key, value in init.kwargs.items():
            setattr(self, key, value)

        deps = init.dependencies

        if self.skip:
            return

        gateway.expose_port(7880, self.host, True)
        gateway.expose_port_to(7882, self.host, 7880, False)

        self.initialized = True

    def retrieve_access(self):
        return LivekitCredentials(
            **{
                "api_key": self.api_key,
                "api_secret": self.api_secret,
                "api_url": f"http://{self.host}:7880",
            }
        )

    def build(self, context: ExecutionContext):
        if self.skip:
            return
        db_service = {
            "labels": [
                "fakts.service=io.livekit.livekit",
                "fakts.builder=livekitio.livekit",
            ],
            "image": self.image,
            "command": self.command,
            "ports": [
                f"{self.port_range[0]}-{self.port_range[1]}:{self.port_range[0]}-{self.port_range[1]}",
                "7881:7881",
            ],
        }

        context.docker_compose.set_nested("services", self.host, db_service)

    def get_options(self):
        with_host = Option(
            subcommand="host",
            help="The fakts url for connection",
            default=self.host,
        )
        with_skip = Option(
            subcommand="skip",
            help="The fakts_next url for connection",
            default=False,
            type=bool,
        )

        return [
            with_host,
            with_skip,
        ]

    def __str__(self) -> str:
        return (
            f"LiveKitBlok(host={self.host}, command={self.command}, image={self.image})"
        )



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass


@dataclass
class LokCredentials:
    issuer: str
    key_type: str
    public_key: str


@service("live.arkitekt.lok")
class LokService(Protocol):
    def retrieve_credentials(self) -> LokCredentials: ...

    def retrieve_labels(self, service_name: str, builder_name: str) -> list[str]: ...

    def register_scopes(self, scopes_dict: Dict[str, str]): ...



from pydantic import BaseModel
from typing import Dict, Any
from arkitekt_next.bloks.services.mount import MountService
from blok import blok, InitContext, Option, ExecutionContext

from blok.tree import YamlFile


@blok(MountService, description="Mounting files into the file tree for use")
class MountBlok:
    def __init__(self) -> None:
        self.mount_path = "mounts"
        self.registered_configs = {}

    def preflight(self, mount_path: str):
        self.mount_path = mount_path

    def build(self, ex: ExecutionContext):
        for name, file in self.registered_configs.items():
            ex.file_tree.set_nested(*f"{self.mount_path}/{name}".split("/"), file)

    def register_mount(self, name: str, file: YamlFile) -> str:
        self.registered_configs[name] = file
        return f"./{self.mount_path}/" + name

    def get_options(self):
        config_path = Option(
            subcommand="mount_path",
            help="Which path to use for configs",
            default=self.mount_path,
            show_default=True,
        )

        return [config_path]



"""Models for ArkitektNext. Thiese include extensiosn for the Fakts Manifest and the User model."""

from hashlib import sha256
import json
from pydantic import BaseModel, Field, field_validator, ConfigDict
from typing import List, Optional


class Requirement(BaseModel):
    key: str
    service: str
    """ The service is the service that will be used to fill the key, it will be used to find the correct instance. It needs to fullfill
    the reverse domain naming scheme"""
    optional: bool = False
    """ The optional flag indicates if the requirement is optional or not. Users should be able to use the client even if the requirement is not met. """
    description: Optional[str] = None
    """ The description is a human readable description of the requirement. Will be show to the user when asking for the requirement."""


class Manifest(BaseModel):
    """A manifest for an app that can be installed in ArkitektNext

    Manifests are used to describe apps that can be installed in ArkitektNext.
    They provide information about the app, such as the
    its globally unique identifier, the version, the scopes it needs, etc.

    This Manifest is send to the Fakts server on initial app configuration,
    and is used to register the app with the Fakts server, which in turn
    will prompt the user to grant the app access to establish itself as
    an ArkitektNext app (and therefore as an OAuth2 client) (see more in the
    Fakts documentation).

    """

    version: str
    """ The version of the app TODO: Should this be a semver? """
    identifier: str
    """ The globally unique identifier of the app: TODO: Should we check for a reverse domain name? """
    scopes: List[str]
    """ Scopes that this app should request from the user """
    logo: Optional[str]
    """ A URL to the logo of the app TODO: We should enforce this to be a http URL as local paths won't work """
    requirements: Optional[List[Requirement]] = Field(default_factory=list)
    """ Requirements that this app has TODO: What are the requirements? """
    model_config = ConfigDict(extra="forbid")

    def hash(self):
        """Hash the manifest

        A manifest describes all the  metadata of an app. This method
        hashes the manifest to create a unique hash for the current configuration of the app.
        This hash can be used to check if the app has changed since the last time it was run,
        and can be used to invalidate caches.

        Returns:
            str: The hash of the manifest

        """

        unsorted_dict = self.model_dump()

        # sort the requirements
        unsorted_dict["requirements"] = sorted(
            unsorted_dict["requirements"], key=lambda x: x["key"]
        )
        # sort the scopes
        unsorted_dict["scopes"] = sorted(unsorted_dict["scopes"])

        # JSON encode the dictionary
        json_dd = json.dumps(unsorted_dict, sort_keys=True)
        # Hash the JSON encoded dictionary
        return sha256(json_dd.encode()).hexdigest()

    @field_validator("identifier")
    def check_identifier(cls, v):
        assert "/" not in v, "The identifier should not contain a /"
        assert len(v) > 0, "The identifier should not be empty"
        assert len(v) < 256, "The identifier should not be longer than 256 characters"
        return v


class User(BaseModel):
    """A user of ArkitektNext

    This model represents a user on ArkitektNext. As herre_next is acgnostic to the
    user model, we need to provide a model that can be used to represent
    the ArkitektNext user. This model is used by the
    :class:`herre_next.fakts.fakts_endpoint_fetcher.FaktsUserFetcher` to
    fetch the user from the associated ArkitektNext Lok instance. This model
    is closely mimicking the OIDC user model, and is therefore compatible
    to represent OIDC users.

    """

    id: str = Field(alias="sub")
    """ The user's id (in lok, this is the user's sub(ject) ID)"""

    username: str = Field(alias="preferred_username")
    """ The user's preferred username """
    email: str = Field(alias="email")
    """ The user's preferred username """



from typing import Optional

from fakts_next.fakts import Fakts
from fakts_next.grants.remote import RemoteGrant
from fakts_next.grants.remote.discovery.well_known import WellKnownDiscovery
from fakts_next.grants.remote import RemoteGrant
from fakts_next.grants.remote.demanders.static import StaticDemander
from fakts_next.grants.remote.demanders.device_code import DeviceCodeDemander
from fakts_next.grants.remote.claimers.post import ClaimEndpointClaimer
from fakts_next.grants.remote.demanders.redeem import RedeemDemander
from fakts_next.cache.file import FileCache
from arkitekt_next.base_models import Manifest


class ArkitektNextFaktsQt(Fakts):
    grant: RemoteGrant


class ArkitektNextFaktsNext(Fakts):
    pass


def build_arkitekt_next_fakts_next(
    manifest: Manifest,
    url: Optional[str] = None,
    no_cache: bool = False,
    headless: bool = False,
    client_kind: str = "development",
) -> ArkitektNextFaktsNext:
    identifier = manifest.identifier
    version = manifest.version

    demander = DeviceCodeDemander(
        manifest=manifest,
        redirect_uri="http://127.0.0.1:6767",
        open_browser=not headless,
        requested_client_kind=client_kind,
    )

    return ArkitektNextFaktsNext(
        grant=RemoteGrant(
            demander=demander,
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
            skip_cache=no_cache,
        ),
    )


def build_arkitekt_next_redeem_fakts_next(
    manifest: Manifest,
    redeem_token: str,
    url,
    no_cache: Optional[bool] = False,
    headless=False,
):
    identifier = manifest.identifier
    version = manifest.version

    return ArkitektNextFaktsNext(
        grant=RemoteGrant(
            demander=RedeemDemander(token=redeem_token, manifest=manifest),
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
        ),
    )


def build_arkitekt_next_token_fakts_next(
    manifest: Manifest,
    token: str,
    url,
):
    identifier = manifest.identifier
    version = manifest.version

    return ArkitektNextFaktsNext(
        grant=RemoteGrant(
            demander=StaticDemander(token=token),
            discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=FileCache(
            cache_file=f".arkitekt_next/cache/{identifier}-{version}_fakts_cache.json",
            hash=manifest.hash() + url,
        ),
    )



from arkitekt_next.bloks.services.certer import CerterService
from arkitekt_next.bloks.services.name import NameService
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
from typing import Dict, Any, Optional

from blok import blok, InitContext
from blok.bloks.services.dns import DnsService

DEFAULT_PUBLIC_URLS = ["127.0.0.1"]
DEFAULT_PUBLIC_HOSTS = ["localhost"]


class ExposedHost(BaseModel):
    host: str
    port: int
    stip_prefix: bool = True


class ExpostedToHost(BaseModel):
    port: int
    host: str
    to: str
    tls: bool = False


class ExposedPort(BaseModel):
    port: int
    host: str
    tls: bool = False
    to: int


@blok("live.arkitekt.certer", description="A self-signed certificate generator")
class SelfSignedBlok:
    def __init__(self) -> None:
        self.host = "certer"
        self.certer_image = "jhnnsrs/certer:next"
        self.certs_mount = "./certs"

    def retrieve_certs_mount(self):
        return self.certs_mount

    def retrieve_depends_on(self):
        return ["certer"]

    def preflight(self, init: InitContext, dns: DnsService):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.public_ips = dns.get_dns_result().ip_addresses
        self.public_hosts = dns.get_dns_result().hostnames

    def build(
        self,
        context: ExecutionContext,
    ):

        context.file_tree.set_nested("certs", {})

        certer_container = {
            "image": self.certer_image,
            "volumes": ["./certs:/certs"],
            "environment": {
                "DOMAIN_NAMES": ",".join(self.public_hosts),
                "IP_ADDRESSED": ",".join(self.public_ips),
            },
        }

        context.docker_compose.set_nested("services", self.host, certer_container)

    def get_options(self):
        with_public_urls = Option(
            subcommand="public_url",
            help="Which public urls to use",
            type=str,
            multiple=True,
            default=DEFAULT_PUBLIC_URLS,
            show_default=True,
        )
        with_public_services = Option(
            subcommand="public_hosts",
            help="Which public hosts to use",
            type=str,
            multiple=True,
            default=DEFAULT_PUBLIC_HOSTS,
            show_default=True,
        )

        return [with_public_urls, with_public_services]



import rich_click as click
from arkitekt_next.cli.options import *
import asyncio
from arkitekt_next.cli.ui import construct_run_panel
from importlib import import_module
from arkitekt_next.cli.utils import import_builder
from rekuest_next.api.schema import NodeKind, BindsInput
from rich.table import Table
from rich.console import Console
from typing import Dict, Any


async def call_app(
    console: Console,
    app: App,
    template_string: str,
    arg: Dict[str, Any],
):
    async with app:
        raise NotImplementedError("Not implemented yet")


@click.command("prod")
@click.option(
    "--url",
    help="The fakts_next url for connection",
    default=DEFAULT_ARKITEKT_URL,
    envvar="FAKTS_URL",
)
@with_builder
@with_token
@with_instance_id
@with_headless
@with_log_level
@with_skip_cache
@click.pass_context
@click.option(
    "--arg",
    "-a",
    "args",
    help="Key Value pairs for the setup",
    type=(str, str),
    multiple=True,
)
@click.option(
    "--template",
    "-t",
    "template",
    help="The template to run",
    type=str,
)
def local(
    ctx,
    entrypoint=None,
    builder=None,
    args=None,
    template: str = None,
    **builder_kwargs,
):
    """Runs the app in production mode

    \n
    You can specify the builder to use with the --builder flag. By default, the easy builder is used, which is designed to be easy to use and to get started with.

    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)
    entrypoint = entrypoint or manifest.entrypoint

    kwargs = dict(args or [])

    builder = import_builder(builder)

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    app = builder(
        **manifest.to_builder_dict(),
        **builder_kwargs,
    )

    panel = construct_run_panel(app)
    console.print(panel)

    asyncio.run(call_app(console, app, template, kwargs))



import rich_click as click
from rich.table import Table
from rich.panel import Panel
from rich.console import Group
from arkitekt_next.cli.vars import *
from arkitekt_next.cli.constants import *
from arkitekt_next.cli.io import write_manifest


@click.group("scopes")
@click.pass_context
def scopes_group(ctx):
    """Inspect, add and remove scopes to this arkitekt_next app

    Scopes are rights that are granted to any arkitekt_next application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt_next platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    pass


@scopes_group.command("add")
@click.argument(
    "SCOPE",
    nargs=-1,
    type=click.Choice(compile_scopes()),
)
@click.pass_context
def add_scopes(ctx, scope):
    """ "Acd scopes

    Scopes are rights that are granted to any arkitekt_next application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt_next platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    if not scope:
        raise click.ClickException("Please provide at least one scope")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if scope:
        manifest.scopes = set(list(scope) + manifest.scopes)
        write_manifest(manifest)
        console.print(f"Scopes Updated to {manifest.scopes}")


@scopes_group.command("remove")
@click.argument(
    "SCOPE",
    nargs=-1,
    type=click.Choice(compile_scopes()),
)
@click.pass_context
def remove_scopes(ctx, scope):
    """Remove scopes

    Scopes are rights that are granted to any arkitekt_next application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt_next platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """
    if not scope:
        raise click.ClickException("Please provide at least one scope to remove")

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    if scope:
        manifest.scopes = set(manifest.scopes) - set(scope)
        write_manifest(manifest)
        console.print(f"Scopes Updated to {manifest.scopes}")


@scopes_group.command("list")
@click.pass_context
def list_scopes(ctx):
    """List all the [i] currently [/] active scopes

    Scopes are rights that are granted to any arkitekt_next application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt_next platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    table = Table.grid()
    table.padding = (0, 1)
    table.add_column("Scope")
    table.add_column("Description")
    for scope in manifest.scopes:
        table.add_row(scope, "TODO: This should be a description")

    panel = Panel(
        Group("[bold green]Demanded Scopes[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)


@scopes_group.command("available")
@click.pass_context
def list_available(ctx):
    """List all the [i] available [/]  scopes

    Scopes are rights that are granted to any arkitekt_next application, and correspond
    to rights that enable feature when interacting with the platform. These scopes
    provide another element of access control to the arkitekt_next platform, and describe
    on top of the users rights, what the application is allowed to do.

    For more information, please visit the [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]


    """

    console = get_console(ctx)

    table = Table.grid()
    table.padding = (0, 1)
    table.add_column("Scope")
    table.add_column("Description")
    for scope in compile_scopes():
        table.add_row(scope, "TODO")

    panel = Panel(
        Group("[bold green]Available Scopes[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    console.print(panel)



import contextvars
from functools import wraps
from pydantic import BaseModel, Field
from herre_next import Herre
from fakts_next import Fakts
from .base_models import Manifest, Requirement
from typing import Callable, Dict, Optional, Protocol, TypeVar, overload
from typing import runtime_checkable

Params = Dict[str, str]


current_service_registry = contextvars.ContextVar(
    "current_service_registry", default=None
)
GLOBAL_SERVICE_REGISTRY = None


def get_default_service_registry():
    global GLOBAL_SERVICE_REGISTRY
    if GLOBAL_SERVICE_REGISTRY is None:
        GLOBAL_SERVICE_REGISTRY = ServiceBuilderRegistry()
    return GLOBAL_SERVICE_REGISTRY


def get_current_service_registry(allow_global=True):
    return current_service_registry.get(get_default_service_registry())


class Registration(BaseModel):
    name: str
    requirement: Requirement
    builder: Callable[[Herre, Fakts, Params], object]
    schema_loader: Callable[[str], object]


@runtime_checkable
class ArkitektService(Protocol):

    def get_service_name(self):
        pass

    def build_service(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        pass

    def get_requirements(self):
        pass

    def get_graphql_schema(self):
        pass

    def get_turms_project(self):
        pass


class BaseArkitektService:

    def get_service_name(self):
        raise NotImplementedError("get_service_name not implemented")

    def build_service(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        raise NotImplementedError("build_service not implemented")

    def get_requirements(self):
        raise NotImplementedError("get_requirements not implemented")

    def get_graphql_schema(self):
        return None

    def get_turms_project(self):
        return None


basic_requirements = [
    Requirement(
        key="lok",
        service="live.arkitekt.lok",
        description="An instance of ArkitektNext Lok to authenticate the user",
    )
]


class ServiceBuilderRegistry:
    def __init__(self, import_services=True):
        self.service_builders: Dict[str, ArkitektService] = {}
        self.additional_requirements: Dict[str, Requirement] = {}

    def register(
        self,
        service: ArkitektService,
    ):

        name = service.get_service_name()

        if name not in self.service_builders:
            self.service_builders[name] = service
        else:
            raise ValueError(f"Service {name} already registered")

    def register_requirement(self, requirement: Requirement):
        if requirement.key in self.additional_requirements:
            raise ValueError(f"Requirement {requirement.key} already registered)")
        self.additional_requirements[requirement.key] = requirement

    def get(self, name):
        return self.services.get(name)

    def build_service_map(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        potentially_needed_services = {
            name: service.build_service(fakts, herre, params, manifest)
            for name, service in self.service_builders.items()
        }

        return {
            key: value
            for key, value in potentially_needed_services.items()
            if value is not None
        }

    def get_requirements(self):

        requirements = [
            Requirement(
                key="lok",
                service="live.arkitekt.lok",
                description="An instance of ArkitektNext Lok to authenticate the user",
            )
        ]
        taken_requirements = set()

        for service in self.service_builders.values():
            for requirement in service.get_requirements():
                if requirement.key not in taken_requirements:
                    taken_requirements.add(requirement.key)
                    requirements.append(requirement)

        for requirement in self.additional_requirements.values():
            if requirement.key not in taken_requirements:
                taken_requirements.add(requirement.key)
                requirements.append(requirement)

        sorted_requirements = sorted(requirements, key=lambda x: x.key)

        return sorted_requirements


class SetupInfo:
    services: Dict[str, object]


import os
import importlib.util
import pkgutil
import traceback
import logging

T = TypeVar("T")


@overload
def require(
    key: str,
    service: str = None,
    description: str = None,
) -> Callable[[T], T]: ...


def require(
    key: str,
    service: str = None,
    description: str = None,
    service_registry: Optional[ServiceBuilderRegistry] = None,
):
    """Register a requirement with the service registry"""
    service_hook_registry = service_registry or get_current_service_registry()

    requirement = Requirement(key=key, service=service, description=description)
    service_hook_registry.register_requirement(requirement)

    return requirement



import sys
import rich_click as click
from arkitekt_next.cli.vars import get_console, get_manifest
import os
from rich.panel import Panel
import subprocess
import uuid
from .io import generate_build
from click import Context
from .types import Flavour, InspectionInput
import yaml
from typing import Dict, Optional
import json
from arkitekt_next.base_models import Requirement
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL
from arkitekt_next.utils import create_arkitekt_next_folder
from rekuest_next.api.schema import TemplateInput


class InspectionError(Exception):
    pass


def build_flavour(flavour_name: str, flavour: Flavour) -> str:
    """Builds the flavour to docker

    Parameters
    ----------
    flavour : Flavour
        The flavour to build
    manifest : Manifest
        The manifest of the app

    Returns
    -------

    tag: str
        The tag of the built docker container

    """

    build_id = str(uuid.uuid4())

    relative_dir = ".arkitekt_next/flavours/{}/".format(flavour_name)

    command = flavour.generate_build_command(build_id, relative_dir)

    docker_run = subprocess.run(" ".join(command), shell=True)

    if docker_run.returncode != 0:
        raise click.ClickException("Could not build docker container")

    return build_id


def inspect_docker_container(build_id: str) -> InspectionInput:
    try:
        # Run 'docker inspect' with the container ID or name
        result = subprocess.run(
            ["docker", "inspect", build_id],
            stdout=subprocess.PIPE,
            text=True,
            check=True,
        )

        # Parse the JSON output
        try:
            container_info = json.loads(result.stdout)
        except json.decoder.JSONDecodeError as e:
            combined_error = result.stdout + result.stderr
            raise InspectionError(
                f"Could not decode JSON output of docker inspect. {combined_error}"
            ) from e

        # Extract size information
        try:
            size = container_info[0][
                "Size"
            ]  # Size of files that have been written to the filesystem
            size_root_fs = container_info[0][
                "Size"
            ]  # Total size of all the files in the container
        except IndexError as e:
            raise InspectionError("Container not found or invalid JSON output") from e
        except KeyError as e:
            raise InspectionError(
                "Size information not found in the container details"
            ) from e

        return size, size_root_fs
    except subprocess.CalledProcessError as e:
        combined_error = e.stdout + e.stderr
        raise InspectionError(f"An error occurred: {e.stdout + e.stderr}") from e


def inspect_templates(build_id: str, url: str) -> list[TemplateInput]:
    try:
        # Run 'docker inspect' with the container ID or name
        process = subprocess.Popen(
            " ".join(
                [
                    "docker",
                    "run",
                    "-it",
                    "--network",
                    "host",
                    build_id,
                    "arkitekt-next",
                    "inspect",
                    "templates",
                    "-mr",
                    "--url",
                    url,
                ]
            ),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        lines = []
        # Poll process for new output until finished
        while True:
            if process.poll() is not None:
                break
            nextline = process.stdout.readline()

            lines.append(nextline.decode("utf-8"))
            sys.stdout.buffer.write(nextline)
            sys.stdout.flush()

        output = process.communicate()[0]
        exitCode = process.returncode

        result = "\n".join(lines)

        # Parse the JSON output
        correct_part = result.split("--START_TEMPLATES--")[1].split(
            "--END_TEMPLATES--"
        )[0]

        try:
            output = json.loads(correct_part)
        except json.decoder.JSONDecodeError as e:
            combined_error = result.stdout + result.stderr
            raise InspectionError(
                f"Could not decode JSON output of docker inspect. {combined_error}"
            ) from e

        return output
    except subprocess.CalledProcessError as e:
        combined_error = e.stdout + e.stderr

        if "No such command" in combined_error:
            raise InspectionError(
                "Could not find the command `arkitekt_next inspect definitions` in the container. Maybe"
                + "you forgot to install arkitekt_next in the container? "
            )

        raise InspectionError(f"An error occurred: {e.stdout + e.stderr}") from e


def inspect_requirements(build_id: str) -> Dict[str, Requirement]:
    try:
        # Run 'docker inspect' with the container ID or name
        result = subprocess.run(
            [
                "docker",
                "run",
                build_id,
                "arkitekt-next",
                "inspect",
                "requirements",
                "-mr",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            text=True,
        )

        # Parse the JSON output
        correct_part = result.stdout.split("--START_REQUIREMENTS--")[1].split(
            "--END_REQUIREMENTS--"
        )[0]

        try:
            output = json.loads(correct_part)
        except json.decoder.JSONDecodeError as e:
            combined_error = result.stdout + result.stderr
            raise InspectionError(
                f"Could not decode JSON output of docker inspect. {combined_error}"
            ) from e

        return output
    except subprocess.CalledProcessError as e:
        combined_error = e.stdout + e.stderr

        if "No such command" in combined_error:
            raise InspectionError(
                "Could not find the command `arkitekt_next inspect definitions` in the container. Maybe"
                + "you forgot to install arkitekt_next in the container? "
            )

        raise InspectionError(f"An error occurred: {e.stdout + e.stderr}") from e


def inspect_build(build_id: str, url: str) -> InspectionInput:
    size, size_root_fs = inspect_docker_container(build_id)
    templates = inspect_templates(build_id, url)
    requirements = inspect_requirements(build_id)

    return InspectionInput(size=size, templates=templates, requirements=requirements)


def get_flavours(ctx: Context, select: Optional[str] = None) -> Dict[str, Flavour]:
    """Gets the flavours for this app"""

    arkitekt_next_folder = create_arkitekt_next_folder()

    flavours_folder = os.path.join(arkitekt_next_folder, "flavours")

    if not os.path.exists(flavours_folder):
        raise click.ClickException(
            "We could not find the flavours folder. Please run `arkitekt_next port init` first to create a buildable flavour"
        )

    flavours = {}

    for dir_name in os.listdir(flavours_folder):
        dir = os.path.join(flavours_folder, dir_name)
        if os.path.isdir(dir):
            if select is not None and select != dir_name:
                continue

            if os.path.exists(os.path.join(dir, "config.yaml")):
                with open(os.path.join(dir, "config.yaml")) as f:
                    valued = yaml.load(f, Loader=yaml.SafeLoader)
                try:
                    flavour = Flavour.model_validate(valued)
                    flavour.check_relative_paths(dir)
                    flavours[dir_name] = flavour

                except Exception as e:
                    get_console(ctx).print_exception()
                    raise click.ClickException(
                        f"Could not load flavour {dir_name} from {dir} ` config.yaml ` is invalid"
                    ) from e

    return flavours


@click.command()
@click.option(
    "--flavour",
    "-f",
    help="The flavour to build. By default all flavours are being built",
    default=None,
    required=False,
)
@click.option(
    "--no-inspect",
    "-n",
    help="Should we skip the inspection of the app?",
    is_flag=True,
    default=False,
)
@click.option(
    "--tag",
    "-t",
    help="Tag the build with a specific tag",
    type=str,
    default=None,
    required=False,
)
@click.option(
    "--url",
    "-u",
    help="The fakts_next server to use",
    type=str,
    default=DEFAULT_ARKITEKT_URL,
)
@click.pass_context
def build(
    ctx: Context,
    flavour: str,
    no_inspect: bool,
    tag: str = None,
    url: str = DEFAULT_ARKITEKT_URL,
) -> None:
    """Builds the arkitekt_next app to docker"""

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    flavours = get_flavours(ctx, select=flavour)

    md = Panel(
        "Starting to Build Containers for App [bold]{}[/bold]".format(
            manifest.identifier
        ),
        subtitle="Selected Flavours: {}".format(", ".join(flavours.keys())),
    )
    console.print(md)

    build_run = str(uuid.uuid4())

    for key, flavour in flavours.items():
        md = Panel(
            "Building Flavour [bold]{}[/bold]".format(key),
            subtitle="This may take a while...",
            subtitle_align="right",
        )
        console.print(md)

        build_tag = build_flavour(key, flavour)

        if tag:
            subprocess.run(["docker", "tag", build_tag, tag], check=True)

        inspection = None
        if not no_inspect:
            inspection = inspect_build(build_tag, url)

        generate_build(build_run, build_tag, key, flavour, manifest, inspection)

        md = Panel(
            "Built Flavour [bold]{}[/bold]".format(key),
            subtitle="Build ID: {}".format(build_run),
            subtitle_align="right",
        )

        console.print(md)



from fakts_next.fakts import Fakts
from typing import Optional
from fakts_next.grants.remote import RemoteGrant
from fakts_next.grants.remote.demanders.device_code import DeviceCodeDemander

from fakts_next.grants.remote.claimers.post import ClaimEndpointClaimer
from fakts_next.grants.remote.discovery.qt.selectable_beacon import (
    SelectBeaconWidget,
    QtSelectableDiscovery,
)
from arkitekt_next.base_models import Manifest
from qtpy import QtCore, QtWidgets
from fakts_next.cache.qt.settings import QtSettingsCache


class ArkitektNextFaktsQtRemoteGrant(RemoteGrant):
    """An ArkitektNext Fakts grant that uses Qt widgets for token and endpoint storage"""

    discovery: QtSelectableDiscovery


class ArkitektNextFaktsQt(Fakts):
    """A Fakts that uses Qt widgets for token and endpoint storage"""

    grant: ArkitektNextFaktsQtRemoteGrant


def build_arkitekt_next_qt_fakts_next(
    manifest: Manifest,
    no_cache: Optional[bool] = False,
    beacon_widget: Optional[QtWidgets.QWidget] = None,
    parent: Optional[QtWidgets.QWidget] = None,
    settings: Optional[QtCore.QSettings] = None,
) -> ArkitektNextFaktsQt:
    beacon_widget = beacon_widget or SelectBeaconWidget(
        parent=parent, settings=settings
    )

    return ArkitektNextFaktsQt(
        grant=ArkitektNextFaktsQtRemoteGrant(
            demander=DeviceCodeDemander(
                manifest=manifest,
                redirect_uri="http://127.0.0.1:6767",
                open_browser=True,
                requested_client_kind="desktop",
            ),
            discovery=QtSelectableDiscovery(
                widget=beacon_widget,
                settings=settings,
                allow_appending_slash=True,
                auto_protocols=["http", "https"],
                additional_beacons=["http://localhost"],
            ),
            claimer=ClaimEndpointClaimer(),
        ),
        cache=QtSettingsCache(settings=settings),
    )



from pydantic import BaseModel
from typing import Dict, Any
from blok import blok, InitContext, Option
from arkitekt_next.bloks.services.admin import AdminService, AdminCredentials


@blok(AdminService, description="sdmin credentials for the application")
class AdminBlok:
    def __init__(self) -> None:
        self.password = "admin"
        self.username = "admin"
        self.email = "admin@admin.com"

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

    def retrieve(self):
        return AdminCredentials(
            password=self.password,
            username=self.username,
            email=self.email,
        )

    def get_options(self):
        with_username = Option(
            subcommand="username",
            help="Which admin username to use",
            default=self.username,
            show_default=True,
        )
        with_username = Option(
            subcommand="password",
            help="Which password to use",
            default=self.password,
            show_default=True,
        )
        with_email = Option(
            subcommand="password",
            help="Which password to use",
            default=self.password,
            show_default=True,
        )

        return [with_username, with_username, with_email]



from typing import Dict, Any, List, Protocol, Optional
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass


@dataclass
class S3Credentials:
    name: str
    access_key: str
    buckets: Dict[str, str]
    host: str
    port: int
    secret_key: str
    protocol: str
    dependency: Optional[str] = None


@service("live.arkitekt.s3")
class S3Service(Protocol):
    def create_buckets(self, buckets: List[str]) -> S3Credentials: ...



import contextvars
from functools import wraps
from pydantic import BaseModel, Field
from herre_next import Herre
from fakts_next import Fakts
from .base_models import Manifest, Requirement
from typing import Callable, Dict, Optional, Protocol, TypeVar, overload
import importlib
import sys
import os
import traceback
import logging
import pkgutil
from typing import runtime_checkable, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from arkitekt_next.apps import App
else:
    App = Any

Params = Dict[str, str]


current_init_hook_registry = contextvars.ContextVar(
    "current_init_hook_registry", default=None
)
GLOBAL_INIT_HOOK_REGISTRY = None


def get_default_init_hook_registry():
    global GLOBAL_INIT_HOOK_REGISTRY
    if GLOBAL_INIT_HOOK_REGISTRY is None:
        GLOBAL_INIT_HOOK_REGISTRY = InitHookRegisty()
    return GLOBAL_INIT_HOOK_REGISTRY


def get_current_init_hook_registry(allow_global=True):
    return current_init_hook_registry.get(get_default_init_hook_registry())


class Registration(BaseModel):
    name: str
    requirement: Requirement
    builder: Callable[[Herre, Fakts, Params], object]
    schema_loader: Callable[[str], object]


@runtime_checkable
class ArkitektService(Protocol):

    def get_service_name(self):
        pass

    def build_service(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        pass

    def get_requirements(self):
        pass

    def get_graphql_schema(self):
        pass

    def get_turms_project(self):
        pass


class BaseArkitektService:

    def get_service_name(self):
        raise NotImplementedError("get_service_name not implemented")

    def build_service(
        self, fakts: Fakts, herre: Herre, params: Params, manifest: Manifest
    ):
        raise NotImplementedError("build_service not implemented")

    def get_requirements(self):
        raise NotImplementedError("get_requirements not implemented")

    def get_graphql_schema(self):
        return None

    def get_turms_project(self):
        return None


basic_requirements = [
    Requirement(
        key="lok",
        service="live.arkitekt.lok",
        description="An instance of ArkitektNext Lok to authenticate the user",
    )
]


InitHook = Callable[[App], None]


class InitHookRegisty:

    def __init__(self):
        self.init_hooks: Dict[str, InitHook] = {}

    def register(
        self,
        function: InitHook,
        name: Optional[str] = None,
        only_cli: bool = False,
    ):

        if name is None:
            name = function.__name__

        if name not in self.init_hooks:
            self.init_hooks[name] = function
        else:
            raise ValueError(f"Service {name} already registered")

    def run_all(self, app: App):
        for hook in self.init_hooks.values():
            hook(app)


T = TypeVar("T")


@overload
def init(
    function_or_actor: T,
) -> T: ...


@overload
def init(
    function_or_actor: None = None,
) -> Callable[[T], T]: ...


def init(
    *func,
    only_cli: bool = False,
    init_hook_registry: InitHookRegisty = None,
):
    """Register a function as an init hook. This function will be called when the app is initialized."""
    init_hook_registry = init_hook_registry or get_current_init_hook_registry()

    if len(func) > 1:
        raise ValueError("You can only register one function or actor at a time.")
    if len(func) == 1:
        function_or_actor = func[0]

        @wraps(function_or_actor)
        def wrapped_function(*args, **kwargs):
            return function_or_actor(*args, **kwargs)

        init_hook_registry.register(wrapped_function)

        wrapped_function.__is_init_hook__ = True

        return wrapped_function

    else:

        def real_decorator(function_or_actor):
            # Simple bypass for now
            @wraps(function_or_actor)
            def wrapped_function(*args, **kwargs):
                return function_or_actor(*args, **kwargs)

            init_hook_registry.register(wrapped_function, only_cli=only_cli)

            wrapped_function.__is_init_hook__ = True

            return wrapped_function

        return real_decorator



from typing import Any, Dict, Optional

from pydantic import BaseModel

from arkitekt_next.bloks.kraph import KraphBlok
from arkitekt_next.bloks.tailscale import TailscaleBlok
from blok import InitContext, Panel, Renderer, blok

from .fluss import FlussBlok
from .internal_docker import InternalDockerBlok
from .kabinet import KabinetBlok
from .livekit import LocalLiveKitBlok
from .mikro import MikroBlok
from .orkestrator import OrkestratorBlok
from .rekuest import RekuestBlok
from .ollama import OllamaBlok


class AdminCredentials(BaseModel):
    password: str
    username: str
    email: str


@blok(
    "live.arkitekt",
    dependencies=[
        MikroBlok.as_dependency(True, True),
        KabinetBlok.as_dependency(True, True),
        RekuestBlok.as_dependency(True, True),
        FlussBlok.as_dependency(True, True),
        InternalDockerBlok.as_dependency(True, True),
        LocalLiveKitBlok.as_dependency(True, True),
        OrkestratorBlok.as_dependency(True, False),
        KraphBlok.as_dependency(True, False),
        TailscaleBlok.as_dependency(True, False),
        OllamaBlok.as_dependency(True, False),
    ],
)
class ArkitektBlok:

    def entry(self, renderer: Renderer):
        renderer.render(
            Panel(
                f"""This is the arkitekt build that allows you to setup a full stack arkitekt application. You can use this to setup a full stack application with the following services""",
                expand=False,
                title="Welcome to Arkitekt!",
                style="bold magenta",
            )
        )

    def build(self, cwd):
        pass



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass


@dataclass
class AdminCredentials:
    password: str
    username: str
    email: str


@service("live.arkitekt.admin")
class AdminService(Protocol):

    def retrieve(self):
        return AdminCredentials(
            password=self.password,
            username=self.username,
            email=self.email,
        )
